#!/usr/bin/env python

'''
Created on Aug 12, 2012
'''

from __future__ import division

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyue@mit.edu"
__date__ = "Aug 12, 2012"

import argparse
import itertools

from pymatgen.io.smartio import read_structure
from pymatgen.analysis.structure_fitter import StructureFitter

parser = argparse.ArgumentParser(description="""
Convenient structure comparison.""", epilog="""
Author: {}
Version: {}
Last updated: {}""".format(__author__, __version__, __date__))
parser.add_argument("structure_files", metavar="structure_files", type=str,
                    nargs='+',
                    help="Names of files to compare")

args = parser.parse_args()
structure_files = args.structure_files
try:
    structures = [(filename, read_structure(filename))
                  for filename in structure_files]
except Exception as ex:
    print "Error reading files. Are they in the right format?"
    print str(ex)

try:
    for (s1, s2) in itertools.combinations(structures, 2):
        fitter = StructureFitter(s1[1], s2[1])
        print "{} ({}) fits {} ({}) = {}".format(s1[0], s1[1].composition,
                                                 s2[0], s2[1].composition,
                                                 fitter.fit_found)
except Exception as ex:
    print "Error fitting."
    print str(ex)



#!/usr/bin/env python

"""
This module provides classes to interface with the Materials Project http REST
interface to enable the creation of data structures and pymatgen objects using
Materials Project data.

To make use of the Materials Project REST interface, you need to be a
registered user of the MaterialsProject, and apply on the MaterialsProject site
for an API key. Currently, the http interface is still in limited beta.
"""

from __future__ import division

__author__ = "Shyue Ping Ong, Shreyas Cholia"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyue@mit.edu"
__date__ = "Jun 8, 2012"

import urllib
import httplib
import json

from pymatgen import Composition, PMGJSONDecoder
from pymatgen.entries.compatibility import MaterialsProjectCompatibility
from pymatgen.entries.exp_entries import ExpEntry


class MPRester(object):
    """
    A class to conveniently interface with the Materials Project REST
    interface.
    """

    supported_properties = ("structure", "initial_structure",
                            "final_structure", "energy", "energy_per_atom",
                            "formation_energy_per_atom", "nsites", "formula",
                            "pretty_formula", "is_hubbard", "elements",
                            "nelements", "e_above_hull", "hubbards",
                            "is_compatible", "entry")

    def __init__(self, api_key, host="www.materialsproject.org"):
        """
        Args:
            api_key:
                A String API key for accessing the MaterialsProject REST
                interface. Please apply on the Materials Project website for
                one.
            host:
                Url of host to access the MaterialsProject REST interface.
                Defaults to the standard Materials Project REST address, but
                can be changed to other urls implementing a similar interface.
        """
        self.host = host
        self.api_key = api_key

    def get_data(self, chemsys_formula_id, data_type="vasp", prop=""):
        """
        Flexible method to get any data using the Materials Project REST
        interface. Generally used by other methods for more specific queries.

        Format of REST return is *always* a list of dict (regardless of the
        number of pieces of data returned. The general format is as follows:

        [{"material_id": material_id, "property_name" : value}, ...]

        Args:
            chemsys_formula_id:
                A chemical system (e.g., Li-Fe-O), or formula (e.g., Fe2O3) or
                materials_id (e.g., 1234).
            data_type:
                Type of data to return. Currently can either be "vasp" or
                "exp".
            prop:
                Property to be obtained. Should be one of the
                MPRestAdaptor.supported_properties. Leave as empty string for a
                general list of useful properties.
        """
        headers = {"x-api-key": self.api_key}
        conn = httplib.HTTPSConnection(self.host)
        if prop:
            url = "/rest/v1/materials/{}/{}/{}".format(chemsys_formula_id,
                                                       data_type, prop)
        else:
            url = "/rest/v1/materials/{}/{}".format(chemsys_formula_id,
                                                       data_type)

        try:
            conn.request("GET", url, headers=headers)
            response = conn.getresponse()
            data = json.loads(response.read(), cls=PMGJSONDecoder)
            if data["valid_response"]:
                return data["response"]
            else:
                raise MPRestError(data["error"])
        except httplib.HTTPException as ex:
            raise MPRestError(str(ex))

    def get_structure_by_material_id(self, material_id, final=True):
        """
        Get a Structure corresponding to a material_id.

        Args:
            material_id:
                Materials Project material_id (an int).
            final:
                Whether to get the final structure, or the initial
                (pre-relaxation) structure. Defaults to True.

        Returns:
            Structure object.
        """
        prop = "final_structure" if final else "initial_structure"
        data = self.get_data(material_id, prop=prop)
        return data[0][prop]

    def get_entry_by_material_id(self, material_id):
        """
        Get a ComputedEntry corresponding to a material_id.

        Args:
            material_id:
                Materials Project material_id (an int).

        Returns:
            ComputedEntry object.
        """
        data = self.get_data(material_id, prop="entry")
        return data[0]["entry"]

    def get_dos_by_material_id(self, material_id):
        """
        Get a ComputedEntry corresponding to a material_id.

        Args:
            material_id:
                Materials Project material_id (an int).

        Returns:
            A Dos object.
        """
        data = self.get_data(material_id, prop="dos")
        return data[0]["dos"]

    def get_bandstructure_by_material_id(self, material_id):
        """
        Get a ComputedEntry corresponding to a material_id.

        Args:
            material_id:
                Materials Project material_id (an int).

        Returns:
            A Bandstructure object.
        """
        data = self.get_data(material_id, prop="bandstructure")
        return data[0]["bandstructure"]

    def get_entries_in_chemsys(self, elements, compatible_only=True):
        """
        Get a list of ComputedEntries in a chemical system. For example,
        elements = ["Li", "Fe", "O"] will return a list of all entries in the
        Li-Fe-O chemical system, i.e., all LixOy, FexOy, LixFey, LixFeyOz, Li,
        Fe and O phases. Extremely useful for creating phase diagrams of entire
        chemical systems.

        Args:
            elements:
                List of element symbols, e.g., ["Li", "Fe", "O"].
            compatible_only:
                Whether to return only "compatible" entries. Compatible entries
                are entries that have been processed using the
                MaterialsProjectCompatibility class, which performs adjustments
                to allow mixing of GGA and GGA+U calculations for more accurate
                phase diagrams and reaction energies.

        Returns:
            List of ComputedEntries.
        """
        data = self.get_data("-".join(elements), prop="entry")
        entries = [d["entry"] for d in data]
        if compatible_only:
            entries = MaterialsProjectCompatibility().process_entries(entries)
        return entries

    def get_exp_thermo_data(self, formula):
        """
        Get a list of ThermoData objects associated with a formula using the
        Materials Project REST interface.

        Args:
            formula:
                A formula to search for.

        Returns:
            List of ThermoData objects.
        """
        return self.get_data(formula, data_type="exp")

    def get_exp_entry(self, formula):
        """
        Returns an ExpEntry object, which is the experimental equivalent of a
        ComputedEntry and can be used for analyses using experimental data.

        Args:
            formula:
                A formula to search for.

        Returns:
            An ExpEntry object.
        """

        return ExpEntry(Composition(formula),
                        self.get_exp_thermo_data(formula))

    def mpquery(self, criteria, properties):
        """
        Performs an advanced mpquery, which is a Mongo-like syntax for directly
        querying the Materials Project database via the mpquery rest interface.
        Please refer to the Moogle advanced help on the mpquery language and
        supported criteria and properties. Essentially, any supported
        properties within MPRestAdaptor should be supported in mpquery.

        Mpquery allows an advanced developer to perform queries which are
        otherwise too cumbersome to perform using the standard convenience
        methods.

        Args:
            criteria:
                Criteria of the query as a mongo-style dict. For example,
                {"elements":{"$in":["Li", "Na", "K"], "$all": ["O"]},
                "nelements":2} selects all Li, Na and K oxides
            properties:
                Properties to request for as a list. For example,
                ["formula", "formation_energy_per_atom"] returns the formula
                and formation energy per atom.

        Returns:
            List of dict of data.
        """
        params = urllib.urlencode({"criteria": criteria,
                                   "properties": properties})
        headers = {"x-api-key": self.api_key}
        conn = httplib.HTTPSConnection(self.host)
        conn.request("POST", "/rest/v1/mpquery", body=params, headers=headers)
        response = conn.getresponse()
        data = json.loads(response.read(), cls=PMGJSONDecoder)
        return data


class MPRestError(Exception):
    """
    Exception class for MPRestAdaptor.
    Raised when the query has problems, e.g., bad query format.
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return "Materials Project REST Error : " + self.msg



#!/usr/bin/env python

"""
This module defines useful physical constants and conversion factors.
All units are in SI units except for conversion factors.
"""

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "1.0"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyue@mit.edu"
__status__ = "Production"
__date__ = "Sep 23, 2011"


"""
Constants. Note that some of these may replicate functionality in
scipy.constants. However, given the difficulty in installing scipy on many
systems, the replication of these constants minimizes scipy dependency.
"""
ELECTRON_CHARGE = 1.602176565e-19
EPSILON_0 = 8.85418781762e-12
BOLTZMANN_CONST = 1.3806488e-23
ELECTRON_VOLT = 1.602176565e-19
AVOGADROS_CONST = 6.02214129e23

"""
Conversion factors
"""
EV_PER_ATOM_TO_J_PER_MOL = ELECTRON_VOLT * AVOGADROS_CONST
EV_PER_ATOM_TO_KJ_PER_MOL = EV_PER_ATOM_TO_J_PER_MOL / 1000
ELECTRON_TO_AMPERE_HOURS = EV_PER_ATOM_TO_J_PER_MOL / 3600
AMU_TO_KG = 1.660538921e-27



#!/usr/bin/env python

"""
Wrapper classes for Cif input and output from Structures.
"""

from __future__ import division

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "1.0"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyue@mit.edu"
__status__ = "Production"
__date__ = "Sep 23, 2011"


import re
import StringIO
import math
import warnings
from collections import OrderedDict

import CifFile
import numpy as np

from pymatgen.core.periodic_table import Element, Specie
from pymatgen.util.io_utils import zopen
from pymatgen.core.lattice import Lattice
from pymatgen.core.structure import Structure
from pymatgen.core.composition import Composition
from pymatgen.core.operations import SymmOp


class CifParser(object):
    """
    A wrapper class around PyCifRW to read Cif and convert into a pymatgen
    Structure object.
    """

    def __init__(self, filename, occupancy_tolerance=1.):
        """
        Args:
            filename:
                Cif filename. bzipped or gzipped cifs are fine too.
            occupancy_tolerance:
                If total occupancy of a site is between 1 and
                occupancy_tolerance, the occupancies will be scaled down to 1.
        """
        self._occupancy_tolerance = occupancy_tolerance
        if isinstance(filename, basestring):
            with zopen(filename, "r") as f:
                self._cif = CifFile.ReadCif(f)
        else:
            self._cif = CifFile.ReadCif(filename)

    @staticmethod
    def from_string(cif_string, occupancy_tolerance=1.):
        output = StringIO.StringIO()
        output.write(cif_string)
        output.seek(0)
        return CifParser(output, occupancy_tolerance)

    def _unique_coords(self, coord_in, primitive, lattice, primlattice):
        """
        Generate unique coordinates using coord and symmetry positions.
        """
        coords = list()
        for op in self.symmetry_operations:
            coord = op.operate(coord_in)
            if primitive:
                cart = lattice.get_cartesian_coords(np.array(coord))
                coord = primlattice.get_fractional_coords(cart)
            coord = np.array([i - math.floor(i) for i in coord])
            if not coord_in_list(coord, coords, 1e-3):
                coords.append(coord)
        return coords

    def _get_structure(self, data, primitive):
        """
        Generate structure from part of the cif.
        """
        spacegroup = data.get("_symmetry_space_group_name_H-M", "P1")

        if len(spacegroup) == 0:
            latt_type = "P"
        else:
            latt_type = spacegroup[0]
        lengths = [float_from_str(data["_cell_length_" + i])
                   for i in ["a", "b", "c"]]
        angles = [float_from_str(data["_cell_angle_" + i])
                  for i in ["alpha", "beta", "gamma"]]
        lattice = Lattice.from_lengths_and_angles(lengths, angles)
        primlattice = lattice.get_primitive_lattice(latt_type)
        try:
            sympos = data["_symmetry_equiv_pos_as_xyz"]
        except:
            try:
                sympos = data["_symmetry_equiv_pos_as_xyz_"]
            except:
                warnings.warn("No _symmetry_equiv_pos_as_xyz type key found. "
                              "Defaulting to P1.")
                sympos
        self.symmetry_operations = parse_symmetry_operations(sympos)

        def parse_symbol(sym):
            m = re.search("([A-Z][a-z]*)", sym)
            if m:
                return m.group(1)
            return ""

        #oxi_states = None
        try:
            oxi_states = dict()
            for i in xrange(len(data["_atom_type_symbol"])):
                oxi_states[data["_atom_type_symbol"][i]] = \
                    float_from_str(data["_atom_type_oxidation_number"][i])
        except:
            oxi_states = None

        coord_to_species = OrderedDict()

        for i in xrange(len(data["_atom_site_type_symbol"])):
            symbol = parse_symbol(data["_atom_site_type_symbol"][i])
            if oxi_states is not None:
                el = Specie(symbol,
                            oxi_states[data["_atom_site_type_symbol"][i]])
            else:
                el = Element(symbol)
            x = float_from_str(data["_atom_site_fract_x"][i])
            y = float_from_str(data["_atom_site_fract_y"][i])
            z = float_from_str(data["_atom_site_fract_z"][i])
            try:
                occu = float_from_str(data["_atom_site_occupancy"][i])
            except:
                occu = 1
            if occu > 0:
                coord = (x, y, z)
                if coord not in coord_to_species:
                    coord_to_species[coord] = {el: occu}
                else:
                    coord_to_species[coord][el] = occu

        allspecies = list()
        allcoords = list()

        for coord, species in coord_to_species.items():
            coords = self._unique_coords(coord, primitive, lattice,
                                         primlattice)
            allcoords.extend(coords)
            allspecies.extend(len(coords) * [species])

        #rescale occupancies if necessary
        for species in allspecies:
            totaloccu = sum(species.values())
            if  1 < totaloccu <= self._occupancy_tolerance:
                for key, value in species.iteritems():
                    species[key] = value / totaloccu

        if primitive:
            struct = Structure(primlattice, allspecies, allcoords)
        else:
            struct = Structure(lattice, allspecies, allcoords)
        return struct.get_sorted_structure()

    def get_structures(self, primitive=True):
        """
        Return list of structures in CIF file. primitive boolean sets whether a
        conventional cell structure or primitive cell structure is returned.

        Args:
            primitive:
                Set to False to return conventional unit cells. Defaults to
                True.

        Returns:
            List of Structures.
        """
        return [self._get_structure(v, primitive)
                for k, v in self._cif.items()]

    @property
    def to_dict(self):
        d = OrderedDict()
        for k, v in self._cif.items():
            d[k] = {}
            for k2, v2 in v.items():
                d[k][k2] = v2
        return d


class CifWriter:
    """
    A wrapper around PyCifRW to write CIF files from pymatgen structures.
    """

    def __init__(self, struct):
        """
        Args:
            struct:
                A pymatgen.core.structure.Structure object.
        """
        block = CifFile.CifBlock()
        latt = struct.lattice
        comp = struct.composition
        no_oxi_comp = Composition.from_formula(comp.formula)
        block["_symmetry_space_group_name_H-M"] = "P 1"
        block["_cell_length_a"] = str(latt.a)
        block["_cell_length_b"] = str(latt.b)
        block["_cell_length_c"] = str(latt.c)
        block["_cell_angle_alpha"] = str(latt.alpha)
        block["_cell_angle_beta"] = str(latt.beta)
        block["_cell_angle_gamma"] = str(latt.gamma)
        block["_chemical_name_systematic"] = "Generated by pymatgen"
        block["_symmetry_Int_Tables_number"] = 1
        block["_chemical_formula_structural"] = str(no_oxi_comp
                                                    .reduced_formula)
        block["_chemical_formula_sum"] = str(no_oxi_comp.formula)
        block["_cell_volume"] = str(latt.volume)

        reduced_comp = Composition.from_formula(no_oxi_comp.reduced_formula)
        el = no_oxi_comp.elements[0]
        amt = comp[el]
        fu = int(amt / reduced_comp[Element(el.symbol)])

        block["_cell_formula_units_Z"] = str(fu)
        block.AddCifItem(([["_symmetry_equiv_pos_site_id",
                            "_symmetry_equiv_pos_as_xyz"]],
                          [[["1"], ["x, y, z"]]]))

        contains_oxidation = True
        symbol_to_oxinum = dict()
        try:
            symbol_to_oxinum = {str(el): el.oxi_state for el in comp.elements}
        except:
            symbol_to_oxinum = {el.symbol: 0 for el in comp.elements}
            contains_oxidation = False
        if contains_oxidation:
            block.AddCifItem(([["_atom_type_symbol",
                                "_atom_type_oxidation_number"]],
                              [[symbol_to_oxinum.keys(),
                                symbol_to_oxinum.values()]]))

        atom_site_type_symbol = []
        atom_site_symmetry_multiplicity = []
        atom_site_fract_x = []
        atom_site_fract_y = []
        atom_site_fract_z = []
        atom_site_attached_hydrogens = []
        atom_site_B_iso_or_equiv = []
        atom_site_label = []
        atom_site_occupancy = []
        count = 1
        for site in struct:
            for sp, occu in site.species_and_occu.items():
                atom_site_type_symbol.append(str(sp))
                atom_site_symmetry_multiplicity.append("1")
                atom_site_fract_x.append(str("{0:f}".format(site.a)))
                atom_site_fract_y.append(str("{0:f}".format(site.b)))
                atom_site_fract_z.append(str("{0:f}".format(site.c)))
                atom_site_attached_hydrogens.append("0")
                atom_site_B_iso_or_equiv.append(".")
                atom_site_label.append(str(sp.symbol) + str(count))
                atom_site_occupancy.append(str(occu))
                count += 1

        block["_atom_site_type_symbol"] = atom_site_type_symbol
        block.AddToLoop("_atom_site_type_symbol",
                        {"_atom_site_label": atom_site_label})
        block.AddToLoop("_atom_site_type_symbol",
                        {"_atom_site_symmetry_multiplicity":
                         atom_site_symmetry_multiplicity})
        block.AddToLoop("_atom_site_type_symbol",
                        {"_atom_site_fract_x": atom_site_fract_x})
        block.AddToLoop("_atom_site_type_symbol",
                        {"_atom_site_fract_y": atom_site_fract_y})
        block.AddToLoop("_atom_site_type_symbol",
                        {"_atom_site_fract_z": atom_site_fract_z})
        block.AddToLoop("_atom_site_type_symbol",
                        {"_atom_site_attached_hydrogens":
                         atom_site_attached_hydrogens})
        block.AddToLoop("_atom_site_type_symbol",
                        {"_atom_site_B_iso_or_equiv":
                         atom_site_B_iso_or_equiv})
        block.AddToLoop("_atom_site_type_symbol",
                        {"_atom_site_occupancy": atom_site_occupancy})

        self._cf = CifFile.CifFile()
        # AJ says: CIF Block names cannot be more than 75 characters or you
        # get an Exception
        self._cf[comp.reduced_formula[0:74]] = block

    def __str__(self):
        """
        Returns the cif as a string.
        """
        return str(self._cf)

    def write_file(self, filename):
        """
        Write the cif file.
        """
        with open(filename, "w") as f:
            f.write(self.__str__())


def around_diff_num(a, b):
    """
    Used to compare differences in fractional coordinates, taking into account
    PBC.
    """
    diff_num = abs(a - b)
    return diff_num if diff_num < 0.5 else abs(1 - diff_num)


def coord_in_list(coord, coord_list, tol):
    """
    Helper method to check if coord is already in a list of coords, subject to
    a tolerance.
    """
    for c in coord_list:
        diff = np.array([around_diff_num(c[i], coord[i]) for i in xrange(3)])
        if (diff < tol).all():
            return True
    return False


def float_from_str(text):
    """
    Remove uncertainty brackets from strings and return the float.
    """
    return float(re.sub("\(.+\)", "", text))


def parse_symmetry_operations(symmops_str):
    ops = []
    for op_str in symmops_str:
        rot_matrix = np.zeros((3, 3))
        trans = np.zeros(3)
        toks = op_str.strip().split(",")
        for i, tok in enumerate(toks):
            for m in re.finditer("([\+\-]*)\s*([x-z\d]+)\/*(\d*)", tok):
                factor = -1 if m.group(1) == "-" else 1
                if m.group(2) in ("x", "y", "z"):
                    j = ord(m.group(2)) - 120
                    rot_matrix[i, j] = factor
                else:
                    num = float(m.group(2))
                    if m.group(3) != "":
                        num = num / float(m.group(3))
                    trans[i] = factor * num
        op = SymmOp.from_rotation_and_translation(rot_matrix, trans)
        ops.append(op)
    return ops



#!/usr/bin/env python

"""
This module defines filters for Transmuter object.
"""

from __future__ import division

__author__ = "Will Richards, Shyue Ping Ong"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "1.0"
__maintainer__ = "Will Richards"
__email__ = "wrichards@mit.edu"
__date__ = "Sep 25, 2012"


from pymatgen.core.periodic_table import smart_element_or_specie
from pymatgen.serializers.json_coders import MSONable
import abc


class AbstractStructureFilter(MSONable):
    """
    Abstract structure filter class.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        pass

    @abc.abstractmethod
    def test(self, structure):
        '''
        Returns a boolean for any structure. Structures that return true are
        kept in the Transmuter object during filtering.
        '''
        return

    @staticmethod
    def from_dict(d):
        for trans_modules in ['filters']:
            mod = __import__('pymatgen.alchemy.' + trans_modules,
                             globals(), locals(), [d['@class']], -1)
            if hasattr(mod, d['@class']):
                trans = getattr(mod, d['@class'])
                return trans(**d['init_args'])
        raise ValueError("Invalid filter dict")


class ContainsSpecieFilter(AbstractStructureFilter):
    """
    Filter for structures containing certain elements or species.
    By default compares by atomic number
    """

    def __init__(self, species, strict_compare=False, AND=True, exclude=False):
        """
        Args:
            species:
                list of species to look for
            AND:
                whether all species must be present to pass (or fail)
                filter.
            strict_compare:
                if true, compares objects by specie or element object
                if false, compares atomic number
            exclude:
                if true, returns false for any structures with the specie
                (excludes them from the Transmuter)
        """
        self._species = map(smart_element_or_specie, species)
        self._strict = strict_compare
        self._AND = AND
        self._exclude = exclude

    def test(self, structure):
        #set up lists to compare
        if not self._strict:
            #compare by atomic number
            atomic_number = lambda x: x.Z
            filter_set = set(map(atomic_number, self._species))
            structure_set = set(map(atomic_number,
                                    structure.composition.elements))
        else:
            #compare by specie or element object
            filter_set = set(self._species)
            structure_set = set(structure.composition.elements)

        if self._AND and filter_set <= structure_set:
            #return true if we aren't excluding since all are in structure
            return not self._exclude
        elif (not self._AND) and filter_set & structure_set:
            #return true if we aren't excluding since one is in structure
            return not self._exclude
        else:
            #return false if we aren't excluding otherwise
            return self._exclude

    def __repr__(self):
        output = ["ContainsSpecieFilter with parameters:"]
        output.append("species = {}".format(self._species))
        output.append("strict_compare = {}".format(self._strict))
        output.append("AND = {}".format(self._AND))
        output.append("exclude = {}".format(self._exclude))
        return  "\n".join(output)

    @property
    def to_dict(self):
        d = {"version": __version__}
        d["@module"] = self.__class__.__module__
        d["@class"] = self.__class__.__name__
        d["init_args"] = {"species": [str(sp) for sp in self._species],
                          "strict_compare": self._strict,
                          "AND": self._AND,
                          "exclude": self._exclude}
        return d


class SpecieProximityFilter(AbstractStructureFilter):
    """
    This filter removes structures that have certain species that are too close
    together.
    """

    def __init__(self, specie_and_min_dist_dict):
        """
        Args:
            specie_and_min_dist_dict:
                A species string to float mapping. For example, {"Na+": 1}
                means that all Na+ ions must be at least 1 Angstrom away from
                each other. Multiple species criteria can be applied. Note that
                the testing is done based on the actual object. If you have a
                structure with Element, you must use {"Na":1} instead to filter
                based on Element and not Specie.
        """
        self.specie_and_min_dist = {smart_element_or_specie(k): v
                                    for k, v
                                    in specie_and_min_dist_dict.items()}

    def test(self, structure):
        all_species = set(self.specie_and_min_dist.keys())
        for site in structure:
            species = site.species_and_occu.keys()
            sp_to_test = set(species).intersection(all_species)
            if sp_to_test:
                max_r = max([self.specie_and_min_dist[sp]
                             for sp in sp_to_test])
                nn = structure.get_neighbors(site, max_r)
                for sp in sp_to_test:
                    for (nnsite, dist) in nn:
                        if sp in nnsite.species_and_occu.keys():
                            if dist < self.specie_and_min_dist[sp]:
                                return False
        return True

    @property
    def to_dict(self):
        d = {"version": __version__}
        d["@module"] = self.__class__.__module__
        d["@class"] = self.__class__.__name__
        d["init_args"] = {"specie_and_min_dist_dict":
                          {str(sp): v
                           for sp, v in self.specie_and_min_dist.items()}}
        return d



__author__ = "Shyue Ping Ong, Anubhav Jain, Michael Kocher, " + \
             "Geoffroy Hautier, William Davidson Richard, Dan Gunter, " + \
             "Shreyas Cholia, Vincent L Chevrier, Rickard Armiento"
__date__ = "Nov 26, 2012"
__version__ = "2.3.0"

#Useful aliases for commonly used objects and modules.

from pymatgen.core.periodic_table import Element, Specie
from pymatgen.core.composition import Composition
from pymatgen.core.structure import Structure, Molecule
from pymatgen.core.lattice import Lattice
from pymatgen.serializers.json_coders import PMGJSONEncoder, PMGJSONDecoder
from pymatgen.electronic_structure.core import Spin, Orbital
from pymatgen.util.io_utils import zopen



from __future__ import division

__author__ = "Will Richards"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "1.2"
__maintainer__ = "Will Richards"
__email__ = "wrichard@mit.edu"
__date__ = "Aug 31, 2012"

from pymatgen.serializers.json_coders import MSONable
from pymatgen.structure_prediction.substitution_probability \
    import SubstitutionProbability

import logging
from operator import mul


class Substitutor(MSONable):

    def __init__(self, threshold=1e-3, **kwargs):
        """
        This substitutor uses the substitution probability class to
        find good substitutions for a given chemistry or structure.

        Args:
            threshold:
                probability threshold for predictions
            kwargs:
                kwargs for the SubstitutionProbability object
                lambda_table, alpha
        """
        self._kwargs = kwargs
        self._sp = SubstitutionProbability(**kwargs)
        self._threshold = threshold

    def pred_from_list(self, species_list):
        """
        There are an exceptionally large number of substitutions to
        look at (260^n), where n is the number of species in the
        list. We need a more efficient than brute force way of going
        through these possibilities. The brute force method would be::

            output = []
            for p in itertools.product(self._sp.species_list
                                       , repeat = len(species_list)):
                if self._sp.conditional_probability_list(p, species_list)
                                       > self._threshold:
                    output.append(dict(zip(species_list,p)))
            return output

        Instead of that we do a branch and bound.

        Args:
            species_list:
                list of species in the starting structure

        Returns:
            list of dictionaries, each including a substitutions
            dictionary, and a probability value
        """
        #calculate the highest probabilities to help us stop the recursion
        max_probabilities = []
        for s2 in species_list:
            max_p = 0
            for s1 in self._sp.species_list:
                max_p = max([self._sp.cond_prob(s1, s2), max_p])
            max_probabilities.append(max_p)
        output = []

        def _recurse(output_prob, output_species):
            best_case_prob = list(max_probabilities)
            best_case_prob[:len(output_prob)] = output_prob
            if reduce(mul, best_case_prob) > self._threshold:
                if len(output_species) == len(species_list):
                    odict = {}
                    odict['substitutions'] = \
                        dict(zip(species_list, output_species))
                    odict['probability'] = reduce(mul, best_case_prob)
                    output.append(odict)
                    return
                for sp in self._sp.species_list:
                    i = len(output_prob)
                    prob = self._sp.cond_prob(sp, species_list[i])
                    _recurse(output_prob + [prob], output_species + [sp])

        _recurse([], [])
        logging.info('{} substitutions found'.format(len(output)))
        return output

    def pred_from_comp(self, composition):
        """
        Similar to pred_from_list except this method returns a list after
        checking that compositions are charge balanced.
        """
        output = []
        predictions = self.pred_from_list(composition.elements)
        for p in predictions:
            subs = p['substitutions']
            charge = 0
            for i_el in composition.elements:
                f_el = subs[i_el]
                charge += f_el.oxi_state * composition[i_el]
            if charge == 0:
                output.append(p)
        logging.info('{} charge balanced '
                     'compositions found'.format(len(output)))
        return output

    @property
    def to_dict(self):
        return {"name": self.__class__.__name__, "version": __version__,
                "kwargs": self._kwargs, "threshold": self._threshold,
                "@module": self.__class__.__module__,
                "@class": self.__class__.__name__}

    @staticmethod
    def from_dict(d):
        t = d['threshold']
        kwargs = d['kwargs']
        return Substitutor(threshold=t, **kwargs)



#!/usr/bin/env python

"""
This module provides classes to perform fitting of two structures.
"""

from __future__ import division

__author__ = "Shyue Ping Ong, Geoffroy Hautier"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "1.0"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyue@mit.edu"
__status__ = "Production"
__date__ = "Jun 29, 2012"

import math
import itertools
import logging
import random
import time
from collections import OrderedDict

import numpy as np

from pymatgen.core.lattice import Lattice
from pymatgen.core.structure import Structure
from pymatgen.core.operations import SymmOp
from pymatgen.core.structure_modifier import StructureEditor
from pymatgen.symmetry.finder import SymmetryFinder

logger = logging.getLogger(__name__)


class StructureFitter(object):
    """
    Class to perform fitting of two structures.

    Attributes:
        fit_found:
            True if a fit was found
        mapping_op:
            Operation that maps the two structures onto one another. None if no
            fit was found.
        structure_a:
            First structure
        structure_b:
            Second structure
    """
    FAST_FIT = 25000
    NORMAL_FIT = 50000
    ACCURATE_FIT = 100000
    EXTREME_FIT = 10000000

    def __init__(self, structure_a, structure_b, tolerance_cell_misfit=0.1,
                 tolerance_atomic_misfit=1.0, supercells_allowed=True,
                 anonymized=False, fitting_accuracy=FAST_FIT,
                 timeout=300, symmetry_tol=0, reduce_structures=True):
        """
        Fits two structures. All fitting parameters have been set with defaults
        that should work in most cases.

        Args:
            structure_a:
                First structure
            structure_b:
                Second structure to try to match with first structure
            tolerance_cell_misfit:
                Tolerance for cell misfit. Default = 0.1
            tolerance_atomic_misfit:
                Tolerance for atomic misfit. Default = 1.0.
            supercells_allowed:
                Whether supercell structures are allowed.  Default = True.
            anonymized:
                Whether to attempt matching of different species.  Setting this
                to true will allow any two structures with the same framework,
                but different species to match to each other. Default = False.
            fitting_accuracy:
                An integer setting for the fitting accuracy.  Corresponds to
                the max number of candidate rotations considered.  Use the
                static variables FAST_FIT, NORMAL_FIT, ACCURATE_FIT and
                EXTREME_FIT to set the tradeoff between accuracy and speed. The
                default FAST_FIT should work reasonably well in most instances.
            timeout:
                Time out in seconds for generating and testing rotations. Note
                that the same timeout will apply to generating and testing
                separately. Defaults to 5 mins, which means a total of 10 mins.
            symmetry_tol:
                If > 0, symmetry checking is performed on the two structures
                based on that symmetry tolerance in angstrom. Structures with
                different symmetries are not fitted to each other. A good value
                is around 0.1. Defaults to 0.
            reduce_structures:
                Whether to reduce structures to the primitive niggli cell.
                Defaults to True.
        """
        self._tolerance_cell_misfit = tolerance_cell_misfit
        self._tolerance_atomic_misfit = tolerance_atomic_misfit
        self._supercells_allowed = supercells_allowed
        self._anonymized = anonymized
        self._max_rotations = fitting_accuracy
        self._timeout = timeout
        # Sort structures so that they have the same arrangement of species.
        self._structure_a = structure_a.get_sorted_structure()
        self._structure_b = structure_b.get_sorted_structure()

        self._start_time = time.time()

        if symmetry_tol:
            finder_a = SymmetryFinder(self._structure_a, symprec=symmetry_tol)
            finder_b = SymmetryFinder(self._structure_b, symprec=symmetry_tol)
            sg_a = finder_a.get_spacegroup_number()
            sg_b = finder_b.get_spacegroup_number()
            same_sg = sg_a == sg_b
            logger.debug("Spacegroup numbers: A-{}, B-{}".format(sg_a, sg_b))
            prim_a = finder_a.find_primitive()
            prim_b = finder_b.find_primitive()
            finder_a = SymmetryFinder(prim_a, symprec=symmetry_tol)
            finder_b = SymmetryFinder(prim_b, symprec=symmetry_tol)
            self._structure_a = finder_a.get_refined_structure()
            self._structure_b = finder_b.get_refined_structure()
            self._supercells_allowed = False

        if reduce_structures:
            self._structure_a = self._structure_a.get_reduced_structure()
            self._structure_b = self._structure_b.get_reduced_structure()

        if not symmetry_tol or same_sg:
            self._mapping_op = None
            if not self._anonymized:
                self.fit(self._structure_a, self._structure_b)
                if self.fit_found:
                    comp = self._structure_a.composition
                    self.el_mapping = {el: el for el in comp.elements}
            else:
                comp_a = structure_a.composition
                comp_b = structure_b.composition
                if len(comp_a.elements) == len(comp_b.elements):
                    el_a = comp_a.elements
                    #Create permutations of the specie/elements in structure A
                    for p in itertools.permutations(el_a):
                        # Create mapping of the specie/elements in structure B
                        # to that of A.
                        # Then create a modified structure with those elements
                        # and try to fit it.
                        el_mapping = dict(zip(comp_b.elements, p))
                        logger.debug("Using specie mapping " + str(el_mapping))
                        mod = StructureEditor(self._structure_b)
                        mod.replace_species(el_mapping)
                        self.fit(self._structure_a, mod.modified_structure)
                        if self._mapping_op != None:
                            #Store successful element mapping
                            self.el_mapping = {el_a: el_b
                                               for el_b, el_a
                                               in el_mapping.items()}
                            break
                else:
                    logger.debug("No. of elements in structures are unequal. "
                                 "Cannot be fitted!")
        elif symmetry_tol:
            self._mapping_op = None
            logger.debug("Symmetry is different.")

    def fit(self, a, b):
        """
        Compares two structures and give the possible affine mapping that
        transforms one into another.
        """
        logger.debug("Structure a")
        logger.debug(str(a))
        logger.debug("Structure b")
        logger.debug(str(b))
        # Check composition first.  If compositions are not the same, do not
        # need to fit further.
        if a.composition.reduced_formula != b.composition.reduced_formula or \
           ((a.num_sites != b.num_sites) and not self._supercells_allowed):
            logger.debug('Compositions do not match')
            return None

        logger.debug('Compositions match')

        # This is cheating, but let's try an identity fit first.
        # In many situations, e.g., when a structure has been topotatically
        # delithiated or substituted, you actually can get a very fast answer
        # without having to try all rotations.
        (mapping_op, biggest_dist, oshift) = self.simple_fit(a, b)

        # If we can't do a simple fit, guess we'd have to do the painful
        # process of finding rotations.
        if not mapping_op:
            # Let's sanitize the structures for fitting. This results in more
            # orthogonal cells, and typically much faster fitting.
            a = a.copy(sanitize=True)
            b = b.copy(sanitize=True)

            # Fitting is done by matching sites in one structure (to_fit)
            # to the other (fixed).
            # We set the structure with fewer sites as fixed,
            # and scale the structures to the same density
            (fixed, to_fit) = self._scale_structures(a, b)

            # Defines the atom misfit tolerance
            tol_atoms = self._tolerance_atomic_misfit * \
                        (3 * 0.7405 * fixed.volume / \
                         (4 * math.pi * fixed.num_sites)) ** (1 / 3)
            logger.debug("Atomic misfit tolerance = %.4f" % (tol_atoms))

            max_sites = float('inf')
            # Determine which type of sites to use for the mapping
            for sp in to_fit.species_and_occu:
                sp_sites = [site for site in to_fit \
                            if site.species_and_occu == sp]
                if len(sp_sites) < max_sites:
                    fit_sites = sp_sites
                    max_sites = len(sp_sites)

            # Set the arbitrary origin
            origin = fit_sites[0]
            logger.debug("Origin = " + str(origin))

            oshift = SymmOp.from_rotation_and_translation(np.eye(3),
                                                          - origin.coords)
            shifted_to_fit = apply_operation(to_fit, oshift)

            #Get candidate rotations
            cand_rot = self._get_candidate_rotations(origin, fixed, to_fit)

            #Reset the clock.
            self._start_time = time.time()
            logger.debug(" Found {} candidate rotations".format(len(cand_rot)))
            if len(cand_rot) == 0:
                logger.debug("No candidate rotations found, returning null. ")
                return None

            # sort the operations, the first ones are the ones with small shear
            # this assures us that we find the smallest cell misfit fits
            sorted_cand_rot = sorted(cand_rot.keys(),
                                     key=lambda r: cand_rot[r])

            for rot in sorted_cand_rot:
                (found_map, mapping_op, biggest_dist) = \
                    self._test_rotation(rot, origin, fixed, shifted_to_fit,
                                        tol_atoms)
                if found_map:
                    break
        else:
            logger.debug("Identity fitting matched!")

        logger.debug("Done testing in {} secs".format(time.time() - \
                                                      self._start_time))

        self._atomic_misfit = biggest_dist / ((3 * 0.7405 * a.volume / \
                                (4 * math.pi * a.num_sites)) ** (1 / 3))

        if mapping_op != None:
            rot = mapping_op.rotation_matrix  # maps to_fit to fixed
            p = sqrt_matrix(np.dot(rot.transpose(), rot))
            scale_matrix = np.eye(3) * self.scale
            newrot = np.dot(scale_matrix, rot)
            # we need to now make sure fitterdata.MappingOp maps b -> a and not
            # the other way around
            mshift = np.dot(rot, oshift.translation_vector)
            finaltranslation = mapping_op.translation_vector + mshift[0]
            composite_op = SymmOp.from_rotation_and_translation(
                                                newrot, finaltranslation)
            self._mapping_op = composite_op if self.fixed_is_a \
                                            else composite_op.inverse
            #self._mapping_op = mapping_op
            self._cell_misfit = shear_invariant(p)

    def simple_fit(self, a, b):
        # Fitting is done by matching sites in one structure (to_fit)
        # to the other (fixed).
        # We set the structure with fewer sites as fixed,
        # and scale the structures to the same density
        (fixed, to_fit) = self._scale_structures(a, b)

        # Defines the atom misfit tolerance
        tol_atoms = self._tolerance_atomic_misfit * (3 * 0.7405 * fixed.volume\
                                  / (4 * math.pi * fixed.num_sites)) ** (1 / 3)
        logger.debug("Atomic misfit tolerance = %.4f" % (tol_atoms))

        max_sites = float('inf')
        # determine which type of sites to use for the mapping
        for sp in to_fit.species_and_occu:
            sp_sites = [site for site in to_fit if site.species_and_occu == sp]
            if len(sp_sites) < max_sites:
                fit_sites = sp_sites
                max_sites = len(sp_sites)

        # Set the arbitrary origin
        origin = fit_sites[0]
        logger.debug("Origin = " + str(origin))

        oshift = SymmOp.from_rotation_and_translation(np.eye(3),
                                                               - origin.coords)
        shifted_to_fit = apply_operation(to_fit, oshift)

        simple_rots = self._get_simple_rotations(fixed, to_fit)
        for rot in simple_rots:
            rot_op = SymmOp.from_rotation_and_translation(rot,
                                                          np.array([0, 0, 0]))
            (found_map, mapping_op, biggest_dist) = self._test_rotation(rot_op,
                                    origin, fixed, shifted_to_fit, tol_atoms)
            if found_map:
                return (mapping_op, biggest_dist, oshift)

        return (None, None, None)

    def _test_rotation(self, rot, origin, fixed, to_fit, tol_atoms):
        tol_atoms_plus = 1.1 * tol_atoms
        found_map = False
        mapping_op = None
        biggest_dist = 0
        logger.debug("Trying candidate rotation : \n" + str(rot))
        for site in fixed:
            if time.time() - self._start_time > self._timeout:
                logger.debug("Timeout reached when testing rotations.")
                break
            if site.species_and_occu == origin.species_and_occu:
                shift = site.coords
                op = SymmOp.from_rotation_and_translation(
                                                    rot.rotation_matrix, shift)
                nstruct = apply_operation(to_fit, op)
                correspondance = OrderedDict()
                all_match = True
                biggest_dist = 0
                # check to see if transformed struct matches fixed structure
                for trans in nstruct:
                    cands = fixed.get_sites_in_sphere(trans.coords,
                                                      tol_atoms_plus)
                    if len(cands) == 0:
                        logger.debug("No candidates found.")
                        all_match = False
                        break
                    cands = sorted(cands, key=lambda a: a[1])
                    (closest, closest_dist) = cands[0]
                    if closest_dist > tol_atoms or \
                       closest.species_and_occu != trans.species_and_occu:
                        logger.debug("Closest dist too large! " + \
                                     "Closest dist = {}".format(closest_dist))
                        all_match = False
                        break
                    correspondance[trans] = closest

                    if closest_dist > biggest_dist:
                        biggest_dist = closest_dist

                if not all_match:
                    continue
                if not are_sites_unique(correspondance.values(), False):
                    all_match = False
                    continue
                else:
                    for k, v in correspondance.items():
                        logger.debug(str(k) + " fits on " + str(v))
                fixed_nb_of_fit_per_site = [0] * len(fixed.sites)
                for t in correspondance:
                    reduced_site = correspondance[t].to_unit_cell
                    for y in range(len(fixed.sites)):
                        if fixed.sites[y] == reduced_site:
                            fixed_nb_of_fit_per_site[y] = \
                                           fixed_nb_of_fit_per_site[y] + 1
                if len(set(fixed_nb_of_fit_per_site)) != 1:
                    all_match = False
                    continue
                # now check to see if the converse is true -- do all of the
                # sites of fixed match up with a site in toFit
                # this used to not be here. This fixes a bug.
                logger.debug("Checking inverse mapping")
                inv_correspondance = OrderedDict()

                # it used to be fixed.getNumSites() != nStruct.getNumSites()
                # so only when the number of sites are different but it's
                # actually better to always check the reverse. This
                # elimininates weird situations where two atoms fit to one
                # (reduced in the unit cell)
                for fixed_site in fixed:
                    cands = nstruct.get_sites_in_sphere(fixed_site.coords,
                                                        tol_atoms_plus)
                    if len(cands) == 0:
                        logger.debug("Rejected because inverse mapping does "
                                     "not fit - Step 1")
                        all_match = False
                        break

                    cands = sorted(cands, key=lambda a: a[1])
                    (closest, closest_dist) = cands[0]

                    if closest_dist > tol_atoms or \
                       closest.species_and_occu != fixed_site.species_and_occu:
                        all_match = False
                        logger.debug("Rejected because inverse mapping does "
                                     "not fit - Step 2")
                        break
                    inv_correspondance[fixed_site] = closest
                if all_match:
                    if not are_sites_unique(inv_correspondance.values(),
                                            False):
                        all_match = False
                        logger.debug("Rejected because two atoms fit to the "
                                     "same site for the inverse")
                        continue

                    self.inv_correspondance = inv_correspondance
                    logger.debug("Correspondance for the inverse")
                    for k, v in inv_correspondance.items():
                        logger.debug("{} fits on {}".format(k, v))

                    # The smallest correspondance array shouldn't have any
                    # equivalent sites
                    if fixed.num_sites != to_fit.num_sites:
                        logger.debug("Testing sites unique")
                        if not are_sites_unique(inv_correspondance.values()):
                            all_match = False
                            logger.debug("Rejected because the smallest "
                                         "correspondance array has equivalent"
                                         " sites.")
                            continue
                    if len(nstruct.sites) == len(fixed.sites):
                        nstruct_nb_of_fit_per_site = [0] * len(nstruct.sites)
                        for t in inv_correspondance:
                            reduced_site = inv_correspondance[t].to_unit_cell
                            for y in range(len(nstruct.sites)):
                                if nstruct.sites[y] == reduced_site:
                                    nstruct_nb_of_fit_per_site[y] = \
                                            nstruct_nb_of_fit_per_site[y] + 1
                        if len(set(nstruct_nb_of_fit_per_site)) != 1:
                            all_match = False
                            continue

                    found_map = True
                    mapping_op = op
                    self.correspondance = correspondance
                    break

        return (found_map, mapping_op, biggest_dist)

    def __str__(self):
        output = ["Fitting structures"]
        output.append("\nStructure 1:")
        output.append(str(self._structure_a))
        output.append("\nStructure 2:")
        output.append(str(self._structure_b))
        output.append("\nFitting parameters:")
        output.append("\tTolerance cell misfit = " + \
                      str(self._tolerance_cell_misfit))
        output.append("\tSupercells allowed = {}".format(self.
                                                         _supercells_allowed))
        output.append("\tAnonymized  = " + str(self._anonymized))
        output.append("\nFitting " + ("succeeded " if self._mapping_op != None\
                                                   else "failed"))
        if self._mapping_op != None:
            output.append("\tMapping op = " + str(self._mapping_op))
            output.append("\tCell misfit = " + str(self._cell_misfit))
            output.append("\tAtomic misfit = " + str(self._atomic_misfit))
            if self._anonymized:
                output.append("Element mapping")
                output.append("\n".join([str(k) + " -> " + str(v) \
                                         for k, v in self.el_mapping.items()]))
        return "\n".join(output)

    def _scale_structures(self, a, b):
        # which structure do we want to fit to the other ?
        # assume that structure b has less sites and switch if needed

        self.fixed_is_a = a.num_sites < b.num_sites
        (fixed, to_fit) = (a, b) if self.fixed_is_a else (b, a)

        # scale the structures to the same density
        rho_a = fixed.num_sites / fixed.volume
        rho_b = to_fit.num_sites / to_fit.volume
        scale = (rho_b / rho_a) ** (1 / 3)
        self.scale = scale
        to_fit = Structure(Lattice(to_fit.lattice.matrix * scale),
                           to_fit.species_and_occu, to_fit.frac_coords)
        return (fixed, to_fit)

    def _get_candidate_rotations(self, origin, fixed, to_fit):
        tol_shear = self._tolerance_cell_misfit
        fixed_basis = fixed.lattice.matrix.transpose()
        # need to generate candidate rotations ...
        lengths = fixed.lattice.abc

        shells = []
        for i in range(3):
            dr = lengths[i] * math.sqrt(tol_shear / 2)
            shell = to_fit.get_neighbors_in_shell(origin.coords, lengths[i],
                                                  dr)
            logger.debug("shell {} radius={} dr={}".format(i, lengths[i], dr))
            shells.append([site for (site, dist) in shell
                           if site.species_and_occu ==
                           origin.species_and_occu])
            logger.debug("No. in shell = {}".format(len(shells[-1])))
        # now generate candidate rotations
        cand_rot = {}  # Dict of SymmOp : float
        a = len(shells[0])
        b = len(shells[1])
        c = len(shells[2])
        total_rots = a * b * c
        if total_rots < self._max_rotations:
            logger.debug("Total rots = {}. Using all rotations.".format(
                                                                   total_rots))
            test_rotations = itertools.product(*shells)
        else:
            logger.warning("Total rots={m} exceed max_rotations={n}.".format(
                                        m=total_rots, n=self._max_rotations))

            def random_rot():
                considered_rots = []
                while len(considered_rots) < self._max_rotations:
                    (x, y, z) = [random.randint(0, i - 1) for i in [a, b, c]]
                    if (x, y, z) not in considered_rots:
                        considered_rots.append((x, y, z))
                        yield (shells[0][x], shells[1][y], shells[2][z])
            test_rotations = random_rot()
        det = np.linalg.det
        inv = np.linalg.inv
        fixed_vol = fixed.volume
        for pool in test_rotations:
            if all([nn.species_and_occu == origin.species_and_occu \
                    for nn in pool]):
                # Can a unitary transformation bring the cell vectors together?
                coords = [nn.coords - origin.coords for nn in pool]
                cell_v = np.array(coords).transpose()
                d = det(cell_v)
                if abs(d) < 0.001 or abs(abs(d) - fixed_vol) > 0.001:
                    continue
                rot = np.dot(fixed_basis, inv(cell_v))
                r = SymmOp.from_rotation_and_translation(rot,
                                                         np.array([0, 0, 0]))

                if r not in cand_rot:
                    transf = r.rotation_matrix
                    transf = np.dot(transf.transpose(), transf)

                    # This resolves a very very strange bug in numpy that
                    # causes random eigenvectors to be returned for matrices
                    # very very close to the identity matrix.
                    transf = np.eye(3) if np.allclose(transf, np.eye(3)) \
                                       else transf
                    pbis = sqrt_matrix(transf)
                    if shear_invariant(pbis) < tol_shear:
                        cand_rot[r] = shear_invariant(pbis)
            if time.time() - self._start_time > self._timeout:
                logger.debug("Timeout reached when generating rotations.")
                break

        return cand_rot

    def _get_simple_rotations(self, fixed, to_fit):
        a1 = fixed.lattice.angles
        a2 = to_fit.lattice.angles
        for i in xrange(3):
            if abs(a1[i] - a2[i]) > 5:
                return []
        fixed_unit_matrix = np.array([row / np.linalg.norm(row) \
                                      for row in fixed.lattice.matrix])
        to_fit_unit_matrix = np.array([row / np.linalg.norm(row) \
                                       for row in to_fit.lattice.matrix])
        possible_rots = []
        inv_fixed = np.linalg.inv(fixed_unit_matrix.transpose())
        #Take into account possible inversion.
        for p in itertools.permutations(to_fit_unit_matrix):
            possible_rots.append(np.dot(np.array(p).transpose(), inv_fixed))
            possible_rots.append(-possible_rots[-1])
        return possible_rots

    @property
    def fit_found(self):
        """
        True if a fit was found.
        """
        return self.mapping_op != None

    @property
    def mapping_op(self):
        """
        Operation mapping the two structures.  None if no fit was found.
        """
        return self._mapping_op

    @property
    def structure_a(self):
        """First input structure"""
        return self._structure_a

    @property
    def structure_b(self):
        """Second input structure"""
        return self._structure_b


def apply_operation(structure, symmop):
    """
    Applies a symmetry operation on a structure.

    Args:
        structure:
            Input structure
        symmop:
            SymmOp to apply

    Returns:
        Modified structure after applying symmop.
    """
    editor = StructureEditor(structure)
    editor.apply_operation(symmop)
    return editor.modified_structure


def sqrt_matrix(matrix):
    """
    Calculates sqrt matrix for input matrix.

    Args:
        matrix:
            Input matrix

    Returns:
        Sqrt matrix.
    """
    d, v = np.linalg.eig(matrix)
    diagonalbis = np.sqrt(d) * np.eye(3)
    return np.dot(v, np.dot(diagonalbis, v.transpose()))


def shear_invariant(matrix):
    """
    Calculates the shear invariant for a matrix.

    Args:
        matrix:
            Input matrix

    Returns:
        Shear invariant
    """
    ans = 0
    for i, j in itertools.combinations(xrange(3), 2):
        ans += (matrix[i][i] - matrix[j][j]) ** 2
        ans += 6 * matrix[i][j] ** 2
    return ans


def are_sites_unique(sites, allow_periodic_image=True):
    """
    Checks if the sites are unique.

    Args:
        sites:
            List of sites to check.
        allow_periodic_image:
            Whether to allow periodic images of sites to map onto one another.

    Returns:
        True if sites are unique, i.e., does not map onto each other.
    """
    for (site1, site2) in itertools.combinations(sites, 2):
        if allow_periodic_image and site1.is_periodic_image(site2):
            return False
        if site1.species_and_occu == site2.species_and_occu and \
           (abs(site1.coords - site2.coords) < 0.1).all():
            return False
    return True



import glob
import os
import sys

from distribute_setup import use_setuptools
use_setuptools(version='0.6.10')
from setuptools import setup, find_packages, Extension

long_description = """
Pymatgen (Python Materials Genomics) is a robust, open-source Python library
for materials analysis. It currently powers the public Materials Project
(http://www.materialsproject.org), an initiative to make calculated properties
on a large number of materials available to materials researchers and
designers. These are some of the main features:
1. Highly flexible classes for the representation of Element, Site, Molecule,
   Structure objects.
2. Extensive io capabilities to manipulate many VASP input and output files
   (http://cms.mpi.univie.ac.at/vasp/) and the crystallographic information
   file format. This includes generating Structure objects from vasp input and
   output. There is also support for Gaussian input files and XYZ file for
   molecules.
3. Comprehensive tool to generate and view compositional and grand canonical
   phase diagrams.
4. Electronic structure analyses (DOS and Bandstructure).
5. Integration with the Materials Project REST API.

The latest stable version of pymatgen and its accompanying documentation
(http://packages.python.org/pymatgen) is always on PyPI. The
bleeding edge developmental version is available at our GitHub repo at
https://github.com/materialsproject/pymatgen. If you wish to
be notified via email of pymatgen releases, you may become a member of
pymatgen's Google Groups page
(https://groups.google.com/forum/?fromgroups#!forum/pymatgen/).
"""

try:
    from numpy.distutils.misc_util import get_numpy_include_dirs
except ImportError:
    print "numpy.distutils.misc_util cannot be imported."
    print "numpy.distutils.misc_util is needed to build the spglib extension."
    print "Please install numpy first before retrying setup."
    sys.exit(-1)

def get_spglib_ext():
    # Get spglib
    spglibs = glob.glob(os.path.join("dependencies", "spglib*"))
    if len(spglibs) == 0:
        raise ValueError("No spglib found in dependencies/")
    spglibdir = spglibs[0]

    # set rest of spglib
    spgsrcdir = os.path.join(spglibdir, "src")
    include_dirs = [spgsrcdir]
    sources = ["cell.c", "debug.c", "hall_symbol.c", "kpoint.c", "lattice.c",
               "mathfunc.c", "pointgroup.c", "primitive.c", "refinement.c",
               "sitesym_database.c", "site_symmetry.c", "spacegroup.c", "spin.c",
               "spg_database.c", "spglib.c", "symmetry.c"]
    sources = [os.path.join(spgsrcdir, srcfile) for srcfile in sources]
    return Extension("pymatgen._spglib",
                     include_dirs=include_dirs + get_numpy_include_dirs(),
                     sources=[os.path.join(spglibdir, "_spglib.c")] + sources)

setup(
    name="pymatgen",
    packages=find_packages(),
    version="2.3.0",
    install_requires=["numpy>=1.5", "pyhull>=1.3.6", "PyCifRW>=3.3"],
    dependency_links = [
        "https://bitbucket.org/jamesrhester/pycifrw/downloads/PyCifRW-3.5.tar.gz"
    ],
    extras_require={"electronic_structure": ["scipy>=0.10"],
                    "plotting": ["matplotlib>=1.1"],
                    "ase_adaptor": ["ase>=3.3"]},
    package_data={"pymatgen.core": ["bond_lengths.json",
                                    "periodic_table.json"],
                  "pymatgen.analysis": ["bvparam_1991.json", "icsd_bv.json"],
                  "pymatgen.io": ["*.cfg"],
                  "pymatgen.entries": ["*.cfg"],
                  "pymatgen.vis": ["ElementColorSchemes.cfg"]},
    author="Shyue Ping Ong, Anubhav Jain, Michael Kocher, Geoffroy Hautier,"
    "Will Richards, Dan Gunter, Shreyas Cholia, Vincent L Chevrier, "
    "Rickard Armiento",
    author_email="shyue@mit.edu, anubhavj@mit.edu, mpkocher@lbnl.gov, "
    "geoffroy.hautier@uclouvain.be, wrichard@mit.edu, dkgunter@lbl.gov, "
    "vincentchevrier@gmail.com, armiento@mit.edu",
    maintainer="Shyue Ping Ong",
    url="https://github.com/materialsproject/pymatgen/",
    license="MIT",
    description="pymatgen is the Python library powering the "
    "Materials Project (www.materialsproject.org).",
    long_description=long_description,
    keywords=["vasp", "gaussian", "materials", "project",
              "electronic", "structure"],
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    download_url="https://github.com/materialsproject/pymatgen/tarball/master",
    ext_modules=[get_spglib_ext()],
    scripts=[os.path.join("scripts", f) for f in os.listdir("scripts")]
)
